# Copyright 2016 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import codecs
from datetime import datetime
import os
from unittest import mock
from unittest.mock import patch

from oslo_utils import uuidutils

from tacker.common import exceptions
from tacker import context
from tacker.db.common_services import common_services_db_plugin
from tacker.db.nfvo import nfvo_db
from tacker.db.nfvo import ns_db
from tacker.db.nfvo import vnffg_db
from tacker.extensions import nfvo
from tacker.manager import TackerManager
from tacker.nfvo.drivers.vim import openstack_driver
from tacker.nfvo import nfvo_plugin
from tacker.plugins.common import constants
from tacker.tests import constants as test_constants
from tacker.tests.unit.db import base as db_base
from tacker.tests.unit.db import utils
from tacker.vnfm import vim_client

SECRET_PASSWORD = '***'
DUMMY_NS_2 = 'ba6bf017-f6f7-45f1-a280-57b073bf78ef'


def dummy_get_vim(*args, **kwargs):
    vim_obj = dict()
    vim_obj['auth_cred'] = utils.get_vim_auth_obj()
    vim_obj['type'] = 'openstack'
    return vim_obj


def _get_template(name):
    filename = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                '../../etc/samples/' + str(name)))
    with codecs.open(filename, encoding='utf-8', errors='strict') as f:
        return f.read()


class FakeDriverManager(mock.Mock):
    def invoke(self, *args, **kwargs):
        if any(x in ['create', 'create_flow_classifier'] for
               x in args):
            return uuidutils.generate_uuid()
        elif 'create_chain' in args:
            return uuidutils.generate_uuid(), uuidutils.generate_uuid()
        elif 'execute_workflow' in args:
            mock_execution = mock.Mock()
            mock_execution.id.return_value = \
                "ba6bf017-f6f7-45f1-a280-57b073bf78ea"
            return mock_execution
        elif ('prepare_and_create_workflow' in args and
              'delete' == kwargs['action'] and
              DUMMY_NS_2 == kwargs['kwargs']['ns']['id']):
            raise nfvo.NoTasksException(action=kwargs['action'],
                                        resource=kwargs['kwargs']['ns']['id'])
        elif ('prepare_and_create_workflow' in args and
              'create' == kwargs['action'] and
              utils.DUMMY_NS_2_NAME == kwargs['kwargs']['ns']['ns']['name']):
            raise nfvo.NoTasksException(
                action=kwargs['action'],
                resource=kwargs['kwargs']['ns']['ns']['name'])


def get_by_name():
    return False


def dummy_get_vim_auth(*args, **kwargs):
    return {'vim_auth': {'username': 'admin', 'password': 'devstack',
                         'project_name': 'nfv', 'user_id': '',
                         'user_domain_name': 'Default',
                         'auth_url': 'http://10.0.4.207/identity/v3',
                         'project_id': '',
                         'project_domain_name': 'Default'},
            'vim_id': '96025dd5-ca16-49f3-9823-958eb04260c4',
            'vim_type': 'openstack', 'vim_name': 'VIM0'}


class FakeClient(mock.Mock):
    def __init__(self, auth):
        pass


class FakeVNFMPlugin(mock.Mock):

    def __init__(self):
        super(FakeVNFMPlugin, self).__init__()
        self.vnf1_vnfd_id = 'eb094833-995e-49f0-a047-dfb56aaf7c4e'
        self.vnf1_vnf_id = '91e32c20-6d1f-47a4-9ba7-08f5e5effe07'
        self.vnf1_update_vnf_id = '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6'
        self.vnf3_vnfd_id = 'e4015e9f-1ef2-49fb-adb6-070791ad3c45'
        self.vnf2_vnfd_id = 'e4015e9f-1ef2-49fb-adb6-070791ad3c45'
        self.vnf3_vnf_id = '7168062e-9fa1-4203-8cb7-f5c99ff3ee1b'
        self.vnf3_update_vnf_id = '10f66bc5-b2f1-45b7-a7cd-6dd6ad0017f5'

        self.cp11_id = 'd18c8bae-898a-4932-bff8-d5eac981a9c9'
        self.cp11_update_id = 'a18c8bae-898a-4932-bff8-d5eac981a9b8'
        self.cp12_id = 'c8906342-3e30-4b2a-9401-a251a7a9b5dd'
        self.cp12_update_id = 'b8906342-3e30-4b2a-9401-a251a7a9b5cc'
        self.cp32_id = '3d1bd2a2-bf0e-44d1-87af-a2c6b2cad3ed'
        self.cp32_update_id = '064c0d99-5a61-4711-9597-2a44dc5da14b'

    def get_vnfd(self, *args, **kwargs):
        if 'VNF1' in args:
            return {'id': self.vnf1_vnfd_id,
                    'name': 'VNF1',
                    'attributes': {'vnfd': _get_template(
                                   'test-nsd-vnfd1.yaml')}}
        elif 'VNF2' in args:
            return {'id': self.vnf3_vnfd_id,
                    'name': 'VNF2',
                    'attributes': {'vnfd': _get_template(
                                   'test-nsd-vnfd2.yaml')}}

    def get_vnfds(self, *args, **kwargs):
        if {'name': ['VNF1']} in args:
            return [{'id': self.vnf1_vnfd_id}]
        elif {'name': ['VNF3']} in args:
            return [{'id': self.vnf3_vnfd_id}]
        else:
            return []

    def get_vnfs(self, *args, **kwargs):
        if {'vnfd_id': [self.vnf1_vnfd_id]} in args:
            return [{'id': self.vnf1_vnf_id}]
        elif {'vnfd_id': [self.vnf3_vnfd_id]} in args:
            return [{'id': self.vnf3_vnf_id}]
        else:
            return None

    def get_vnf(self, *args, **kwargs):
        if self.vnf1_vnf_id in args:
            return self.get_dummy_vnf1()
        elif self.vnf1_update_vnf_id in args:
            return self.get_dummy_vnf1_update()
        elif self.vnf3_vnf_id in args:
            return self.get_dummy_vnf3()
        elif self.vnf3_update_vnf_id in args:
            return self.get_dummy_vnf3_update()

    def get_vnf_resources(self, *args, **kwargs):
        if self.vnf1_vnf_id in args:
            return self.get_dummy_vnf1_details()
        elif self.vnf1_update_vnf_id in args:
            return self.get_dummy_vnf1_update_details()
        elif self.vnf3_vnf_id in args:
            return self.get_dummy_vnf3_details()
        elif self.vnf3_update_vnf_id in args:
            return self.get_dummy_vnf3_update_details()

    def get_dummy_vnf1_details(self):
        return [{'name': 'CP11', 'id': self.cp11_id},
                {'name': 'CP12', 'id': self.cp12_id}]

    def get_dummy_vnf1_update_details(self):
        return [{'name': 'CP11', 'id': self.cp11_update_id},
                {'name': 'CP12', 'id': self.cp12_update_id}]

    def get_dummy_vnf3_details(self):
        return [{'name': 'CP32', 'id': self.cp32_id}]

    def get_dummy_vnf3_update_details(self):
        return [{'name': 'CP32', 'id': self.cp32_update_id}]

    def get_dummy_vnf1(self):
        return {'description': 'dummy_vnf_description',
                'vnfd_id': self.vnf1_vnfd_id,
                'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'name': 'dummy_vnf1',
                'attributes': {}}

    def get_dummy_vnf1_update(self):
        return {'description': 'dummy_vnf_description',
                'vnfd_id': self.vnf1_vnfd_id,
                'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'name': 'dummy_vnf1_update',
                'attributes': {}}

    def get_dummy_vnf3(self):
        return {'description': 'dummy_vnf_description',
                'vnfd_id': self.vnf3_vnfd_id,
                'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'name': 'dummy_vnf2',
                'attributes': {}}

    def get_dummy_vnf3_update(self):
        return {'description': 'dummy_vnf_description',
                'vnfd_id': self.vnf3_vnfd_id,
                'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                'tenant_id': 'ad7ebc56538745a08ef7c5e97f8bd437',
                'name': 'dummy_vnf_update',
                'attributes': {}}

    def _update_vnf_scaling(self, *args, **kwargs):
        pass


class TestNfvoPlugin(db_base.SqlTestCase):
    def setUp(self):
        super(TestNfvoPlugin, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()

    def _mock_driver_manager(self):
        self._driver_manager = mock.Mock(wraps=FakeDriverManager())
        self._driver_manager.__contains__ = mock.Mock(
            return_value=True)
        fake_driver_manager = mock.Mock()
        fake_driver_manager.return_value = self._driver_manager
        self._mock(
            'tacker.common.driver_manager.DriverManager', fake_driver_manager)

    def _insert_dummy_vim(self):
        session = self.context.session
        vim_db = nfvo_db.Vim(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vim',
            description='fake_vim_description',
            type='openstack',
            status='Active',
            deleted_at=datetime.min,
            placement_attr={'regions': ['RegionOne']})
        vim_auth_db = nfvo_db.VimAuth(
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            password='encrypted_pw',
            auth_url='http://localhost/identity',
            vim_project={'name': 'test_project'},
            auth_cred={'username': 'test_user', 'user_domain_id': 'default',
                       'project_domain_id': 'default',
                       'key_type': 'fernet_key'})
        session.add(vim_db)
        session.add(vim_auth_db)
        session.flush()

    def _insert_dummy_vim_barbican(self):
        session = self.context.session
        vim_db = nfvo_db.Vim(
            id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_vim',
            description='fake_vim_description',
            type='openstack',
            status='Active',
            deleted_at=datetime.min,
            placement_attr={'regions': ['RegionOne']})
        vim_auth_db = nfvo_db.VimAuth(
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            password='encrypted_pw',
            auth_url='http://localhost/identity',
            vim_project={'name': 'test_project'},
            auth_cred={'username': 'test_user', 'user_domain_id': 'default',
                       'project_domain_id': 'default',
                       'key_type': 'barbican_key',
                       'secret_uuid': 'fake-secret-uuid'})
        session.add(vim_db)
        session.add(vim_auth_db)
        session.flush()

    def test_create_vim(self):
        vim_dict = utils.get_vim_obj()
        vim_type = 'openstack'
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        res = self.nfvo_plugin.create_vim(self.context, vim_dict)
        self._cos_db_plugin.create_event.assert_any_call(
            self.context, evt_type=constants.RES_EVT_CREATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VIM,
            tstamp=mock.ANY)
        self._driver_manager.invoke.assert_any_call(
            vim_type, 'register_vim', vim_obj=vim_dict['vim'])
        self.assertIsNotNone(res)
        self.assertEqual(SECRET_PASSWORD, res['auth_cred']['password'])
        self.assertIn('id', res)
        self.assertIn('placement_attr', res)
        self.assertIn('created_at', res)
        self.assertIn('updated_at', res)
        self.assertEqual(False, res['is_default'])
        self.assertEqual('openstack', res['type'])

    def test_delete_vim(self):
        self._insert_dummy_vim()
        vim_type = 'openstack'
        vim_id = '6261579e-d6f3-49ad-8bc3-a9cb974778ff'
        self.context.tenant_id = 'ad7ebc56538745a08ef7c5e97f8bd437'
        vim_obj = self.nfvo_plugin._get_vim(self.context, vim_id)
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin.get_auth_dict'
                   ).start()
        mock.patch('tacker.nfvo.workflows.vim_monitor.vim_monitor_utils.'
                   'delete_vim_monitor'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        self.nfvo_plugin.delete_vim(self.context, vim_id)
        self._driver_manager.invoke.assert_called_once_with(
            vim_type, 'deregister_vim',
            vim_obj=vim_obj)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_DELETE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VIM,
            tstamp=mock.ANY)

    def test_update_vim(self):
        vim_dict = {'vim': {'id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                            'vim_project': {'name': 'new_project'},
                            'auth_cred': {'username': 'new_user',
                                          'password': 'new_password'}}}
        vim_type = 'openstack'
        vim_auth_username = vim_dict['vim']['auth_cred']['username']
        vim_project = vim_dict['vim']['vim_project']
        self._insert_dummy_vim()
        self.context.tenant_id = 'ad7ebc56538745a08ef7c5e97f8bd437'
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        res = self.nfvo_plugin.update_vim(self.context, vim_dict['vim']['id'],
                                          vim_dict)
        vim_obj = self.nfvo_plugin._get_vim(
            self.context, vim_dict['vim']['id'])
        vim_obj['updated_at'] = None
        self._driver_manager.invoke.assert_called_with(
            vim_type, 'register_vim',
            vim_obj=vim_obj)
        self.assertIsNotNone(res)
        self.assertIn('id', res)
        self.assertIn('placement_attr', res)
        self.assertEqual(vim_project, res['vim_project'])
        self.assertEqual(vim_auth_username, res['auth_cred']['username'])
        self.assertEqual(SECRET_PASSWORD, res['auth_cred']['password'])
        self.assertIn('updated_at', res)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_UPDATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VIM,
            tstamp=mock.ANY)

    def test_update_vim_barbican(self):
        vim_dict = {'vim': {'id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                            'vim_project': {'name': 'new_project'},
                            'auth_cred': {'username': 'new_user',
                                          'password': 'new_password'}}}
        vim_type = 'openstack'
        vim_auth_username = vim_dict['vim']['auth_cred']['username']
        vim_project = vim_dict['vim']['vim_project']
        self._insert_dummy_vim_barbican()
        self.context.tenant_id = 'ad7ebc56538745a08ef7c5e97f8bd437'
        old_vim_obj = self.nfvo_plugin._get_vim(
            self.context, vim_dict['vim']['id'])
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        res = self.nfvo_plugin.update_vim(self.context, vim_dict['vim']['id'],
                                          vim_dict)
        vim_obj = self.nfvo_plugin._get_vim(
            self.context, vim_dict['vim']['id'])
        vim_obj['updated_at'] = None
        self._driver_manager.invoke.assert_called_with(
            vim_type, 'delete_vim_auth',
            vim_id=vim_obj['id'],
            auth=old_vim_obj['auth_cred'])
        self.assertIsNotNone(res)
        self.assertIn('id', res)
        self.assertIn('placement_attr', res)
        self.assertEqual(vim_project, res['vim_project'])
        self.assertEqual(vim_auth_username, res['auth_cred']['username'])
        self.assertEqual(SECRET_PASSWORD, res['auth_cred']['password'])
        self.assertIn('updated_at', res)
        self._cos_db_plugin.create_event.assert_called_with(
            self.context, evt_type=constants.RES_EVT_UPDATE, res_id=mock.ANY,
            res_state=mock.ANY, res_type=constants.RES_TYPE_VIM,
            tstamp=mock.ANY)

    def _insert_dummy_vnffg_template(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template={'vnffgd': utils.vnffgd_tosca_template},
            template_source='onboarded')
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg_template_inline(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='11da9f20-9347-4283-bc68-eb98061ef8f7',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='dummy_vnffgd_inline',
            description='dummy_vnffgd_description_inline',
            template={'vnffgd': utils.vnffgd_tosca_template},
            template_source='inline')
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg_param_template(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template={'vnffgd': utils.vnffgd_tosca_param_template})
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg_multi_param_template(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template={'vnffgd': utils.vnffgd_tosca_multi_param_template})
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg_no_classifier_template(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template={'vnffgd': utils.vnffgd_tosca_no_classifier_template})
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg_duplicate_criteria_template(self):
        session = self.context.session
        vnffg_template = vnffg_db.VnffgTemplate(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            description='fake_template_description',
            template={'vnffgd': utils.vnffgd_tosca_dupl_criteria_template})
        session.add(vnffg_template)
        session.flush()
        return vnffg_template

    def _insert_dummy_vnffg(self):
        session = self.context.session
        vnffg = vnffg_db.Vnffg(
            id='ffc1a59b-65bb-4874-94d3-84f639e63c74',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='dummy_vnffg',
            description="fake vnffg",
            vnffgd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            status='ACTIVE',
            vnf_mapping={'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effe07',
                         'VNF3': '7168062e-9fa1-4203-8cb7-f5c99ff3ee1b'})
        session.add(vnffg)
        nfp = vnffg_db.VnffgNfp(
            id='768f76a7-9025-4acd-b51c-0da609759983',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            status="ACTIVE",
            name='Forwarding_path1',
            vnffg_id='ffc1a59b-65bb-4874-94d3-84f639e63c74',
            path_id=51,
            symmetrical=False)
        session.add(nfp)
        sfc = vnffg_db.VnffgChain(
            id='f28e33bc-1061-4762-b942-76060bbd59c4',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            symmetrical=False,
            chain=[{'connection_points': [
                'd18c8bae-898a-4932-bff8-d5eac981a9c9',
                'c8906342-3e30-4b2a-9401-a251a7a9b5dd'],
                'name': 'dummy_vnf1'},
                {'connection_points': ['3d1bd2a2-bf0e-44d1-87af-a2c6b2cad3ed'],
                 'name': 'dummy_vnf2'}],
            path_id=51,
            status='ACTIVE',
            nfp_id='768f76a7-9025-4acd-b51c-0da609759983',
            instance_id='bcfb295e-578e-405b-a349-39f06b25598c')
        session.add(sfc)
        fc = vnffg_db.VnffgClassifier(
            id='a85f21b5-f446-43f0-86f4-d83bdc5590ab',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='classifier_one',
            status='ACTIVE',
            instance_id='3007dc2d-30dc-4651-9184-f1e6273cc0b6',
            chain_id='f28e33bc-1061-4762-b942-76060bbd59c4',
            nfp_id='768f76a7-9025-4acd-b51c-0da609759983')
        session.add(fc)
        match = vnffg_db.ACLMatchCriteria(
            id='bdb0f2db-d4c2-42a2-a1df-426079ecc443',
            vnffgc_id='a85f21b5-f446-43f0-86f4-d83bdc5590ab',
            eth_src=None, eth_dst=None, eth_type=None, vlan_id=None,
            vlan_pcp=None, mpls_label=None, mpls_tc=None, ip_dscp=None,
            ip_ecn=None, ip_src_prefix=None, ip_dst_prefix='192.168.1.2/24',
            source_port_min=None, source_port_max=None,
            destination_port_min=80, destination_port_max=1024, ip_proto=6,
            network_id=None, network_src_port_id=None,
            network_dst_port_id=None, tenant_id=None, icmpv4_type=None,
            icmpv4_code=None, arp_op=None, arp_spa=None, arp_tpa=None,
            arp_sha=None, arp_tha=None, ipv6_src=None, ipv6_dst=None,
            ipv6_flabel=None, icmpv6_type=None, icmpv6_code=None,
            ipv6_nd_target=None, ipv6_nd_sll=None, ipv6_nd_tll=None)
        session.add(match)
        session.flush()
        return vnffg

    def test_validate_tosca(self):
        template = utils.vnffgd_tosca_template
        self.nfvo_plugin.validate_tosca(template)

    def test_validate_tosca_missing_tosca_ver(self):
        template = utils.vnffgd_template
        self.assertRaises(nfvo.ToscaParserFailed,
                          self.nfvo_plugin.validate_tosca,
                          template)

    def test_validate_tosca_invalid(self):
        template = utils.vnffgd_invalid_tosca_template
        self.assertRaises(nfvo.ToscaParserFailed,
                          self.nfvo_plugin.validate_tosca,
                          template)

    def test_validate_vnffg_properties(self):
        template = {'vnffgd': utils.vnffgd_tosca_template}
        self.nfvo_plugin.validate_vnffg_properties(template)

    def test_validate_vnffg_properties_wrong_number(self):
        template = {'vnffgd': utils.vnffgd_wrong_cp_number_template}
        self.assertRaises(nfvo.VnffgdWrongEndpointNumber,
                          self.nfvo_plugin.validate_vnffg_properties,
                          template)

    def test_create_vnffgd(self):
        vnffgd_obj = utils.get_dummy_vnffgd_obj()
        result = self.nfvo_plugin.create_vnffgd(self.context, vnffgd_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('template', result)
        self.assertIn('template_source', result)
        self.assertEqual('dummy_vnffgd', result['name'])
        self.assertEqual('dummy_vnffgd_description', result['description'])
        self.assertEqual('onboarded', result['template_source'])

    def test_create_vnffgd_inline(self):
        vnffgd_obj = utils.get_dummy_vnffgd_obj_inline()
        result = self.nfvo_plugin.create_vnffgd(self.context, vnffgd_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('template', result)
        self.assertEqual('inline', result['template_source'])

    def test_create_vnffgd_no_input_description(self):
        vnffgd_obj = utils.get_dummy_vnffgd_obj_no_description()
        result = self.nfvo_plugin.create_vnffgd(self.context, vnffgd_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('template', result)
        self.assertIn('template_source', result)
        self.assertEqual('Example VNFFG template', result['description'])

    def test_create_vnffgd_no_input_name(self):
        vnffgd_obj = utils.get_dummy_vnffgd_obj_no_name()
        result = self.nfvo_plugin.create_vnffgd(self.context, vnffgd_obj)
        self.assertIsNotNone(result)
        self.assertIn('id', result)
        self.assertIn('template', result)
        self.assertIn('template_source', result)
        self.assertEqual('example_vnffgd', result['name'])

    def test_create_vnffg_abstract_types(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg_obj = utils.get_dummy_vnffg_obj()
            result = self.nfvo_plugin.create_vnffg(self.context, vnffg_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertEqual('PENDING_CREATE', result['status'])
            self._driver_manager.invoke.assert_called_with(
                mock.ANY, mock.ANY,
                name=mock.ANY,
                path_id=mock.ANY,
                vnfs=mock.ANY,
                fc_ids=mock.ANY,
                auth_attr=mock.ANY,
                symmetrical=mock.ANY,
                correlation=mock.ANY)

    @mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin.create_vnffgd')
    def test_create_vnffg_abstract_types_inline(self, mock_create_vnffgd):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            mock_create_vnffgd.return_value = {'id':
                    '11da9f20-9347-4283-bc68-eb98061ef8f7'}
            self._insert_dummy_vnffg_template_inline()
            vnffg_obj = utils.get_dummy_vnffg_obj_inline()
            result = self.nfvo_plugin.create_vnffg(self.context, vnffg_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertEqual('PENDING_CREATE', result['status'])
            self.assertEqual('dummy_vnffg_inline', result['name'])
            mock_create_vnffgd.assert_called_once_with(mock.ANY, mock.ANY)
            self._driver_manager.invoke.assert_called_with(
                mock.ANY, mock.ANY,
                name=mock.ANY,
                path_id=mock.ANY,
                vnfs=mock.ANY,
                fc_ids=mock.ANY,
                auth_attr=mock.ANY,
                symmetrical=mock.ANY,
                correlation=mock.ANY)

    def test_create_vnffg_param_values(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_param_template()
            vnffg_obj = utils.get_dummy_vnffg_param_obj()
            result = self.nfvo_plugin.create_vnffg(self.context, vnffg_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertEqual('PENDING_CREATE', result['status'])
            self._driver_manager.invoke.assert_called_with(
                mock.ANY, mock.ANY,
                name=mock.ANY,
                path_id=mock.ANY,
                vnfs=mock.ANY,
                fc_ids=mock.ANY,
                auth_attr=mock.ANY,
                symmetrical=mock.ANY,
                correlation=mock.ANY)

    def test_create_vnffg_no_classifier(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_no_classifier_template()
            vnffg_obj = utils.get_dummy_vnffg_no_classifier_obj()
            result = self.nfvo_plugin.create_vnffg(self.context, vnffg_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertEqual('PENDING_CREATE', result['status'])
            self._driver_manager.invoke.assert_called_with(
                mock.ANY, mock.ANY,
                name=mock.ANY,
                path_id=mock.ANY,
                vnfs=mock.ANY,
                fc_ids=mock.ANY,
                auth_attr=mock.ANY,
                symmetrical=mock.ANY,
                correlation=mock.ANY)

    def test_create_vnffg_param_value_format_error(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            self._insert_dummy_vnffg_param_template()
            vnffg_obj = utils.get_dummy_vnffg_str_param_obj()
            self.assertRaises(nfvo.VnffgParamValueFormatError,
                              self.nfvo_plugin.create_vnffg,
                              self.context, vnffg_obj)

    def test_create_vnffg_template_param_not_parse(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            self._insert_dummy_vnffg_multi_param_template()
            vnffg_obj = utils.get_dummy_vnffg_param_obj()
            self.assertRaises(nfvo.VnffgTemplateParamParsingException,
                              self.nfvo_plugin.create_vnffg,
                              self.context, vnffg_obj)

    def test_create_vnffg_vnf_mapping(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg_obj = utils.get_dummy_vnffg_obj_vnf_mapping()
            result = self.nfvo_plugin.create_vnffg(self.context, vnffg_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertEqual('PENDING_CREATE', result['status'])
            self._driver_manager.invoke.assert_called_with(
                mock.ANY, mock.ANY,
                name=mock.ANY,
                path_id=mock.ANY,
                vnfs=mock.ANY,
                fc_ids=mock.ANY,
                auth_attr=mock.ANY,
                symmetrical=mock.ANY,
                correlation=mock.ANY
            )

    def test_create_vnffg_duplicate_criteria(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_duplicate_criteria_template()
            vnffg_obj = utils.get_dummy_vnffg_obj_dupl_criteria()
            self.assertRaises(nfvo.NfpDuplicatePolicyCriteria,
                              self.nfvo_plugin.create_vnffg,
                              self.context, vnffg_obj)

    def test_update_vnffg_nonexistent_vnf(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_vnf_mapping()
            updated_vnffg['vnffg']['symmetrical'] = True
            updated_vnf_mapping = \
                {'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6',
                 'VNF3': '5c7f5631-9e74-46e8-b3d2-397c0eda9d0b'}
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            self.assertRaises(nfvo.VnffgInvalidMappingException,
                              self.nfvo_plugin.update_vnffg,
                              self.context, vnffg['id'], updated_vnffg)

    def test_update_vnffg_empty_vnf_mapping_dict(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_vnf_mapping()
            updated_vnffg['vnffg']['symmetrical'] = True
            updated_vnf_mapping = {}
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            self.assertRaises(nfvo.VnfMappingNotFoundException,
                              self.nfvo_plugin.update_vnffg,
                              self.context, vnffg['id'], updated_vnffg)

    def test_update_vnffg_vnf_mapping_key_none(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_vnf_mapping()
            updated_vnffg['vnffg']['symmetrical'] = True
            updated_vnf_mapping = None
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            self.assertRaises(nfvo.VnfMappingNotFoundException,
                              self.nfvo_plugin.update_vnffg,
                              self.context, vnffg['id'], updated_vnffg)

    def test_update_vnffg_vnfd_not_in_vnffg_template(self):
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_vnf_mapping()
            updated_vnffg['vnffg']['symmetrical'] = True
            updated_vnf_mapping = \
                {'VNF2': '91e32c20-6d1f-47a4-9ba7-08f5e5effad7',
                 'VNF3': '5c7f5631-9e74-46e8-b3d2-397c0eda9d0b'}
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            self.assertRaises(nfvo.VnfMappingNotValidException,
                              self.nfvo_plugin.update_vnffg,
                              self.context, vnffg['id'], updated_vnffg)

    def test_update_vnffg_vnf_mapping(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_vnf_mapping()
            updated_vnffg['vnffg']['symmetrical'] = True
            expected_mapping = {'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6',
                                'VNF3': '10f66bc5-b2f1-45b7-a7cd-6dd6ad0017f5'}

            updated_vnf_mapping = \
                {'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6',
                 'VNF3': '10f66bc5-b2f1-45b7-a7cd-6dd6ad0017f5'}
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            result = self.nfvo_plugin.update_vnffg(self.context, vnffg['id'],
                                            updated_vnffg)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertIn('vnf_mapping', result)
            self.assertEqual('ffc1a59b-65bb-4874-94d3-84f639e63c74',
                             result['id'])
            self.assertEqual('PENDING_UPDATE', result['status'])
            for vnfd, vnf in result['vnf_mapping'].items():
                self.assertIn(vnfd, expected_mapping)
                self.assertEqual(vnf, expected_mapping[vnfd])
            self._driver_manager.invoke.assert_called_with(mock.ANY,
                                                           mock.ANY,
                                                           vnfs=mock.ANY,
                                                           fc_ids=mock.ANY,
                                                           chain_id=mock.ANY,
                                                           auth_attr=mock.ANY)

    def test_update_vnffg_vnffgd_template(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_update_vnffgd_template()
            expected_mapping = {'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6'}

            updated_vnf_mapping = \
                {'VNF1': '91e32c20-6d1f-47a4-9ba7-08f5e5effaf6'}
            updated_vnffg['vnffg']['vnf_mapping'] = updated_vnf_mapping
            result = self.nfvo_plugin.update_vnffg(self.context, vnffg['id'],
                                            updated_vnffg)
            self.assertIn('id', result)
            self.assertIn('status', result)
            self.assertIn('vnf_mapping', result)
            self.assertEqual('ffc1a59b-65bb-4874-94d3-84f639e63c74',
                             result['id'])
            for vnfd, vnf in result['vnf_mapping'].items():
                self.assertIn(vnfd, expected_mapping)
                self.assertEqual(vnf, expected_mapping[vnfd])
            self._driver_manager.invoke.assert_called_with(mock.ANY,
                                                           mock.ANY,
                                                           vnfs=mock.ANY,
                                                           fc_ids=mock.ANY,
                                                           chain_id=mock.ANY,
                                                           auth_attr=mock.ANY)

    def test_update_vnffg_legacy_vnffgd_template(self):
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock.patch('tacker.common.driver_manager.DriverManager',
                       side_effect=FakeDriverManager()).start()
            self._insert_dummy_vnffg_template()
            vnffg = self._insert_dummy_vnffg()
            updated_vnffg = utils.get_dummy_vnffg_obj_legacy_vnffgd_template()
            self.assertRaises(nfvo.UpdateVnffgException,
                              self.nfvo_plugin.update_vnffg,
                              self.context, vnffg['id'], updated_vnffg)

    def test_delete_vnffg(self):
        self._insert_dummy_vnffg_template()
        vnffg = self._insert_dummy_vnffg()
        self._mock_driver_manager()
        mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin._get_vim_from_vnf',
                   side_effect=dummy_get_vim).start()
        self.nfvo_plugin = nfvo_plugin.NfvoPlugin()
        mock.patch('tacker.db.common_services.common_services_db_plugin.'
                   'CommonServicesPluginDb.create_event'
                   ).start()
        self._cos_db_plugin =\
            common_services_db_plugin.CommonServicesPluginDb()
        self.nfvo_plugin.delete_vnffg(self.context, vnffg['id'])
        self._driver_manager.invoke.assert_called_with(mock.ANY, mock.ANY,
                                                       fc_id=mock.ANY,
                                                       auth_attr=mock.ANY)

    def _insert_dummy_ns_template(self):
        session = self.context.session
        attributes = {
            'nsd': 'imports: [VNF1, VNF2]\ntopology_template:\n  inputs:\n  '
                   '  vl1_name: {default: net_mgmt, description: name of VL1'
                   ' virtuallink, type: string}\n    vl2_name: {default: '
                   'net0, description: name of VL2 virtuallink, type: string'
                   '}\n  node_templates:\n    VL1:\n      properties:\n     '
                   '   network_name: {get_input: vl1_name}\n        vendor: '
                   'tacker\n      type: tosca.nodes.nfv.VL\n    VL2:\n      '
                   'properties:\n        network_name: {get_input: vl2_name}'
                   '\n        vendor: tacker\n      type: tosca.nodes.nfv.VL'
                   '\n    VNF1:\n      requirements:\n      - {virtualLink1: '
                   'VL1}\n      - {virtualLink2: VL2}\n      type: tosca.node'
                   's.nfv.VNF1\n    VNF2: {type: tosca.nodes.nfv.VNF2}\ntosca'
                   '_definitions_version: tosca_simple_profile_for_nfv_1_0_0'
                   '\n'}
        nsd_template = ns_db.NSD(
            id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='fake_template',
            vnfds={'tosca.nodes.nfv.VNF1': 'vnf1',
                   'tosca.nodes.nfv.VNF2': 'vnf2'},
            description='fake_nsd_template_description',
            deleted_at=datetime.min,
            template_source='onboarded')
        session.add(nsd_template)
        for (key, value) in attributes.items():
            attribute_db = ns_db.NSDAttribute(
                id=uuidutils.generate_uuid(),
                nsd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
                key=key,
                value=value)
            session.add(attribute_db)
        session.flush()
        return nsd_template

    def _insert_dummy_ns_template_inline(self):
        session = self.context.session
        attributes = {
            'nsd': 'imports: [VNF1, VNF2]\ntopology_template:\n  inputs:\n  '
                   '  vl1_name: {default: net_mgmt, description: name of VL1'
                   ' virtuallink, type: string}\n    vl2_name: {default: '
                   'net0, description: name of VL2 virtuallink, type: string'
                   '}\n  node_templates:\n    VL1:\n      properties:\n     '
                   '   network_name: {get_input: vl1_name}\n        vendor: '
                   'tacker\n      type: tosca.nodes.nfv.VL\n    VL2:\n      '
                   'properties:\n        network_name: {get_input: vl2_name}'
                   '\n        vendor: tacker\n      type: tosca.nodes.nfv.VL'
                   '\n    VNF1:\n      requirements:\n      - {virtualLink1: '
                   'VL1}\n      - {virtualLink2: VL2}\n      type: tosca.node'
                   's.nfv.VNF1\n    VNF2: {type: tosca.nodes.nfv.VNF2}\ntosca'
                   '_definitions_version: tosca_simple_profile_for_nfv_1_0_0'
                   '\n'}
        nsd_template = ns_db.NSD(
            id='be18005d-5656-4d81-b499-6af4d4d8437f',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            name='dummy_NSD',
            vnfds={'tosca.nodes.nfv.VNF1': 'vnf1',
                   'tosca.nodes.nfv.VNF2': 'vnf2'},
            description='dummy_nsd_description',
            deleted_at=datetime.min,
            template_source='inline')
        session.add(nsd_template)
        for (key, value) in attributes.items():
            attribute_db = ns_db.NSDAttribute(
                id=uuidutils.generate_uuid(),
                nsd_id='be18005d-5656-4d81-b499-6af4d4d8437f',
                key=key,
                value=value)
            session.add(attribute_db)
        session.flush()
        return nsd_template

    def _insert_dummy_ns(self, status='ACTIVE'):
        session = self.context.session
        ns = ns_db.NS(
            id='ba6bf017-f6f7-45f1-a280-57b073bf78ea',
            name='dummy_ns',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            status=status,
            nsd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            description='dummy_ns_description',
            deleted_at=datetime.min)
        session.add(ns)
        session.flush()
        return ns

    def _insert_dummy_ns_2(self):
        session = self.context.session
        ns = ns_db.NS(
            id=DUMMY_NS_2,
            name='fake_ns',
            tenant_id='ad7ebc56538745a08ef7c5e97f8bd437',
            status='ACTIVE',
            nsd_id='eb094833-995e-49f0-a047-dfb56aaf7c4e',
            vim_id='6261579e-d6f3-49ad-8bc3-a9cb974778ff',
            description='fake_ns_description',
            deleted_at=datetime.min)
        session.add(ns)
        session.flush()
        return ns

    def test_create_nsd(self):
        nsd_obj = utils.get_dummy_nsd_obj()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            result = self.nfvo_plugin.create_nsd(self.context, nsd_obj)
            self.assertIsNotNone(result)
            self.assertEqual('dummy_NSD', result['name'])
            self.assertIn('id', result)
            self.assertEqual('dummy_NSD', result['name'])
            self.assertEqual('onboarded', result['template_source'])
            self.assertEqual('8819a1542a5948b68f94d4be0fd50496',
                             result['tenant_id'])
            self.assertIn('attributes', result)
            self.assertIn('created_at', result)
            self.assertIn('updated_at', result)

    def test_create_nsd_without_vnfd_imports(self):
        nsd_obj = utils.get_dummy_nsd_obj()
        # Remove vnfd import section from nsd.
        nsd_obj['nsd']['attributes']['nsd'].pop('imports')
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            self.assertRaises(nfvo.ToscaParserFailed,
                          self.nfvo_plugin.create_nsd, self.context, nsd_obj)

    def test_create_nsd_inline(self):
        nsd_obj = utils.get_dummy_nsd_obj_inline()
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            result = self.nfvo_plugin.create_nsd(self.context, nsd_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertEqual('dummy_NSD_inline', result['name'])
            self.assertEqual('inline', result['template_source'])
            self.assertEqual('8819a1542a5948b68f94d4be0fd50496',
                             result['tenant_id'])
            self.assertIn('attributes', result)
            self.assertIn('created_at', result)
            self.assertIn('updated_at', result)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim',
        return_value={"vim_type": "openstack"})
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    @mock.patch.object(openstack_driver.OpenStack_Driver, 'get_mistral_client')
    @mock.patch.object(uuidutils, 'generate_uuid',
        return_value=test_constants.UUID)
    def test_create_ns(self, mock_uuid, mock_mistral_client,
            mock_get_by_name, mock_get_vimi, mock_auth_dict):
        self._insert_dummy_ns_template()
        self._insert_dummy_vim()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }

        sample_yaml = 'std.create_ns' + test_constants.UUID + ':\n  '\
            'input:\n  - ns\n  output:\n    mgmt_ip_address_VNF1: '\
            '<% $.mgmt_ip_address_VNF1 %>\n    mgmt_ip_address_VNF2: '\
            '<% $.mgmt_ip_address_VNF2 %>\n    status_VNF1: '\
            '<% $.status_VNF1 %>\n    status_VNF2: '\
            '<% $.status_VNF2 %>\n    vim_id_VNF1: <% $.vim_id_VNF1 %>\n'\
            '    vim_id_VNF2: <% $.vim_id_VNF2 %>\n    vnf_id_VNF1: '\
            '<% $.vnf_id_VNF1 %>\n    vnf_id_VNF2: <% $.vnf_id_VNF2 %>\n  '\
            'tasks:\n    create_ns_VNF1:\n      action: tacker.create_vnf '\
            'body=<% $.ns.VNF1 %>\n      input:\n        body: '\
            '<% $.ns.VNF1 %>\n      on-success:\n      '\
            '- wait_vnf_active_VNF1\n      publish:\n        '\
            'mgmt_ip_address_VNF1: <% task(create_ns_VNF1).result.'\
            'vnf.mgmt_ip_address %>\n        status_VNF1: <% '\
            'task(create_ns_VNF1).result.vnf.status %>\n'\
            '        vim_id_VNF1: <% task(create_ns_VNF1).'\
            'result.vnf.vim_id %>\n        vnf_id_VNF1: <% '\
            'task(create_ns_VNF1).result.vnf.id %>\n'\
            '    create_ns_VNF2:\n      action: tacker.create_vnf '\
            'body=<% $.ns.VNF2 %>\n      input:\n        '\
            'body: <% $.ns.VNF2 %>\n      on-success:\n'\
            '      - wait_vnf_active_VNF2\n      publish:\n        '\
            'mgmt_ip_address_VNF2: <% task(create_ns_VNF2).result.vnf.'\
            'mgmt_ip_address %>\n        status_VNF2: <% task(create_ns_VNF2)'\
            '.result.vnf.status %>\n        vim_id_VNF2: <% '\
            'task(create_ns_VNF2).result.vnf.vim_id %>\n'\
            '        vnf_id_VNF2: <% task(create_ns_VNF2).result.vnf.id '\
            '%>\n    delete_vnf_VNF1:\n      action: tacker.delete_vnf vnf'\
            '=<% $.vnf_id_VNF1%>\n      input:\n        body:\n          '\
            'vnf:\n            attributes:\n              force: false\n'\
            '    delete_vnf_VNF2:\n      action: tacker.delete_vnf'\
            ' vnf=<% $.vnf_id_VNF2%>\n      input:\n        body:\n'\
            '          vnf:\n            attributes:\n              '\
            'force: false\n    wait_vnf_active_VNF1:\n      action:'\
            ' tacker.show_vnf vnf=<% $.vnf_id_VNF1 %>\n      on-success:\n'\
            '      - delete_vnf_VNF1: <% $.status_VNF1="ERROR" %>\n      '\
            'publish:\n        mgmt_ip_address_VNF1: \' <% '\
            'task(wait_vnf_active_VNF1).result.vnf.mgmt_ip_address\n'\
            '          %>\'\n        status_VNF1: <% '\
            'task(wait_vnf_active_VNF1).result.vnf.status %>\n      '\
            'retry:\n        break-on: <% $.status_VNF1 = "ERROR"'\
            ' %>\n        continue-on: <% $.status_VNF1 = "PENDING_CREATE" '\
            '%>\n        count: 10\n        delay: 10\n    '\
            'wait_vnf_active_VNF2:\n      action: tacker.show_vnf vnf=<% '\
            '$.vnf_id_VNF2 %>\n      on-success:\n      '\
            '- delete_vnf_VNF2: <% $.status_VNF2="ERROR" %>\n      '\
            'publish:\n        mgmt_ip_address_VNF2: \' <% '\
            'task(wait_vnf_active_VNF2).result.vnf.mgmt_ip_address\n'\
            '          %>\'\n        status_VNF2: '\
            '<% task(wait_vnf_active_VNF2).result.vnf.status %>\n'\
            '      retry:\n        break-on: <% $.status_VNF2 = '\
            '"ERROR" %>\n        continue-on: '\
            '<% $.status_VNF2 = "PENDING_CREATE" %>\n        '\
            'count: 10\n        '\
            'delay: 10\n  type: direct\nversion: \'2.0\'\n'

        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()

            ns_obj = utils.get_dummy_ns_obj()
            result = self.nfvo_plugin.create_ns(self.context, ns_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertEqual(ns_obj['ns']['nsd_id'], result['nsd_id'])
            self.assertEqual(ns_obj['ns']['name'], result['name'])
            self.assertIn('status', result)
            self.assertIn('tenant_id', result)
            mock_mistral_client().workflows.create.\
                assert_called_with(sample_yaml)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim',
        return_value={"vim_type": "openstack"})
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    @mock.patch.object(openstack_driver.OpenStack_Driver, 'get_mistral_client')
    def test_create_ns_empty_description(self, mock_mistral_client,
                                         mock_get_by_name,
                                         mock_get_vimi, mock_auth_dict):
        self._insert_dummy_ns_template()
        self._insert_dummy_vim()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()

            ns_obj = utils.get_dummy_ns_obj()
            ns_obj['ns']['description'] = ''
            result = self.nfvo_plugin.create_ns(self.context, ns_obj)
            self.assertIn('id', result)
            self.assertEqual(ns_obj['ns']['name'], result['name'])
            self.assertEqual('', result['description'])

    @mock.patch('tacker.nfvo.nfvo_plugin.NfvoPlugin.create_nsd')
    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim',
        return_value={"vim_type": "openstack"})
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    @mock.patch.object(openstack_driver.OpenStack_Driver, 'get_mistral_client')
    def test_create_ns_inline(self, mock_mistral_client, mock_get_by_name,
                              mock_get_vimi, mock_auth_dict, mock_create_nsd):
        self._insert_dummy_ns_template_inline()
        self._insert_dummy_vim()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()
            mock_create_nsd.return_value = {'id':
                            'be18005d-5656-4d81-b499-6af4d4d8437f'}

            ns_obj = utils.get_dummy_ns_obj_inline()
            result = self.nfvo_plugin.create_ns(self.context, ns_obj)
            self.assertIsNotNone(result)
            self.assertIn('id', result)
            self.assertEqual(ns_obj['ns']['nsd_id'], result['nsd_id'])
            self.assertEqual(ns_obj['ns']['name'], result['name'])
            self.assertEqual('dummy_ns_inline', result['name'])
            self.assertIn('status', result)
            self.assertIn('tenant_id', result)
            mock_create_nsd.assert_called_once_with(mock.ANY, mock.ANY)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim',
        return_value={"vim_type": "openstack"})
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    @mock.patch('tacker.common.driver_manager.DriverManager.invoke',
        side_effect=nfvo.NoTasksException(action='create',
                                          resource='ns'))
    def test_create_ns_workflow_no_task_exception(
            self, mock_mistral_client, mock_get_by_name,
            mock_get_vimi, mock_auth_dict):
        self._insert_dummy_ns_template()
        self._insert_dummy_vim()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }
        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()

            ns_obj = utils.get_dummy_ns_obj_2()
            self.assertRaises(nfvo.NoTasksException,
                              self.nfvo_plugin.create_ns,
                              self.context, ns_obj)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim')
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    def test_delete_ns(self, mock_get_by_name, mock_get_vim, mock_auth_dict):
        self._insert_dummy_vim()
        self._insert_dummy_ns_template()
        self._insert_dummy_ns()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }

        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()
            result = self.nfvo_plugin.delete_ns(self.context,
                'ba6bf017-f6f7-45f1-a280-57b073bf78ea')
            self.assertIsNotNone(result)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim')
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    @mock.patch("tacker.db.nfvo.ns_db.NSPluginDb.delete_ns_post")
    def test_delete_ns_no_task_exception(
            self, mock_delete_ns_post, mock_get_by_name, mock_get_vim,
            mock_auth_dict):

        self._insert_dummy_vim()
        self._insert_dummy_ns_template()
        self._insert_dummy_ns_2()
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }

        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()
            self.nfvo_plugin.delete_ns(self.context,
                DUMMY_NS_2)
        mock_delete_ns_post.assert_called_with(
            self.context, DUMMY_NS_2, None, None, force_delete=False)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim')
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    def test_delete_ns_force(self, mock_get_by_name,
                             mock_get_vim, mock_auth_dict):
        self._insert_dummy_vim()
        self._insert_dummy_ns_template()
        self._insert_dummy_ns(status='PENDING_DELETE')
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }
        nsattr = {'ns': {'attributes': {'force': True}}}

        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()
            result = self.nfvo_plugin.delete_ns(self.context,
                'ba6bf017-f6f7-45f1-a280-57b073bf78ea', ns=nsattr)
            self.assertIsNotNone(result)

    @mock.patch.object(nfvo_plugin.NfvoPlugin, 'get_auth_dict')
    @mock.patch.object(vim_client.VimClient, 'get_vim')
    @mock.patch.object(nfvo_plugin.NfvoPlugin, '_get_by_name')
    def test_delete_ns_force_non_admin(self, mock_get_by_name,
                                       mock_get_vim, mock_auth_dict):
        self._insert_dummy_vim()
        self._insert_dummy_ns_template()
        self._insert_dummy_ns(status='PENDING_DELETE')
        mock_auth_dict.return_value = {
            'auth_url': 'http://127.0.0.1',
            'token': 'DummyToken',
            'project_domain_name': 'dummy_domain',
            'project_name': 'dummy_project'
        }
        nsattr = {'ns': {'attributes': {'force': True}}}

        with patch.object(TackerManager, 'get_service_plugins') as \
                mock_plugins:
            mock_plugins.return_value = {'VNFM': FakeVNFMPlugin()}
            mock_get_by_name.return_value = get_by_name()
            non_admin_context = context.Context(user_id=None,
                                            tenant_id=None,
                                            is_admin=False)
            self.assertRaises(exceptions.AdminRequired,
                              self.nfvo_plugin.delete_ns,
                              non_admin_context,
                              'ba6bf017-f6f7-45f1-a280-57b073bf78ea',
                              ns=nsattr)



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for vnf lcm APIs.

"""

from tacker.api.validation import parameter_types
from tacker.objects import fields

_extManagedVirtualLinkData = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.identifier,
            'vnfVirtualLinkDescId': parameter_types.identifier_in_vnfd,
            'resourceId': parameter_types.identifier_in_vim,
            'vimConnectionId': parameter_types.identifier
        },
        'required': ['id', 'vnfVirtualLinkDescId', 'resourceId'],
        'additionalProperties': True,
    },
}

_ipaddresses = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'type': {'enum': fields.IpAddressType.ALL},
            'subnetId': parameter_types.identifier_in_vim,
            'fixedAddresses': {'type': 'array'},
            'numDynamicAddresses': parameter_types.positive_integer
        },
        'if': {'properties': {'type': {'const': fields.IpAddressType.IPV4}}},
        'then': {
            'properties': {
                'fixedAddresses': {
                    'type': 'array',
                    'items': {'type': 'string', 'format': 'ipv4'}
                }
            }
        },
        'else': {
            'properties': {
                'fixedAddresses': {
                    'type': 'array',
                    'items': {'type': 'string', 'format': 'ipv6'}
                }
            }
        },
        'required': ['type'],
        'oneOf': [
            {'required': ['numDynamicAddresses']},
            {'required': ['fixedAddresses']}
        ],
        'additionalProperties': True
    }
}

_ipOverEthernetAddressData = {
    'type': 'object',
    'properties': {
        'macAddress': parameter_types.mac_address_or_none,
        'ipAddresses': _ipaddresses,
    },
    'anyOf': [
        {'required': ['macAddress']},
        {'required': ['ipAddresses']}
    ],
    'additionalProperties': True
}

_cpProtocolData = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'layerProtocol': {'type': 'string',
                              'enum': 'IP_OVER_ETHERNET'
                              },
            'ipOverEthernet': _ipOverEthernetAddressData,
        },
        'required': ['layerProtocol'],
        'additionalProperties': True,
    }
}

_vnfExtCpConfig = {
    'type': 'array', 'minItems': 1, 'maxItems': 1,
    'items': {
        'type': 'object',
        'properties': {
            'cpInstanceId': parameter_types.identifier_in_vnf,
            'linkPortId': parameter_types.identifier,
            'cpProtocolData': _cpProtocolData,
        },
        'additionalProperties': True,
    }
}

_vnfExtCpData = {
    'type': 'array', 'minItems': 1,
    'items': {
        'type': 'object',
        'properties': {
            'cpdId': parameter_types.identifier_in_vnfd,
            'cpConfig': _vnfExtCpConfig,
        },
        'required': ['cpdId', 'cpConfig'],
        'additionalProperties': True,
    },
}

_resourceHandle = {
    'type': 'object',
    'properties': {
        'resourceId': parameter_types.identifier_in_vim,
        'vimConnectionId': parameter_types.identifier,
        'vimLevelResourceType': {'type': 'string', 'maxLength': 255},
    },
    'required': ['resourceId'],
    'additionalProperties': True,
}

_extLinkPortData = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.identifier,
            'resourceHandle': _resourceHandle,
        },
        'required': ['id', 'resourceHandle'],
        'additionalProperties': True,
    }
}

_extVirtualLinkData = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.identifier,
            'resourceId': parameter_types.identifier_in_vim,
            'vimConnectionId': parameter_types.identifier,
            'extCps': _vnfExtCpData,
            'extLinkPorts': _extLinkPortData,

        },
        'required': ['id', 'resourceId', 'extCps'],
        'additionalProperties': True,
    }
}

_vimConnectionInfo = {
    'type': 'array',
    'maxItems': 1,
    'items': {
        'type': 'object',
        'properties': {
            'id': parameter_types.identifier,
            'vimId': parameter_types.identifier,
            'vimType': {'type': 'string', 'minLength': 1, 'maxLength': 255},
            'interfaceInfo': parameter_types.keyvalue_pairs,
            'accessInfo': parameter_types.keyvalue_pairs,
        },
        'required': ['id', 'vimType'],
        'additionalProperties': True,
    }
}

_versions = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'vnfSoftwareVersion': {'type': 'string'},
            'vnfdVersions': {
                'type': 'array',
                'items': {'type': 'string'}
            }
        },
        'required': ['vnfSoftwareVersion']
    }
}

_vnf_products = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'vnfProductName': {'type': 'string'},
            'versions': _versions
        },
        'required': ['vnfProductName']
    }
}

_vnf_products_from_providers = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'vnfProvider': {'type': 'string'},
            'vnfProducts': _vnf_products
        },
        'required': ['vnfProvider']
    }
}

_lifecycle_change_notifications_filter = {
    'type': 'object',
    'properties': {
        'vnfInstanceSubscriptionFilter': {
            'type': 'object',
            'properties': {
                'vnfdIds': {
                    'type': 'array',
                    'items': parameter_types.identifier
                },
                'vnfProductsFromProviders': _vnf_products_from_providers,
                'vnfInstanceIds': {
                    'type': 'array',
                    'items': parameter_types.identifier
                },
                'vnfInstanceNames': {
                    'type': 'array',
                    'items': {'type': 'string'}
                }
            }
        },
        'notificationTypes': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'VnfLcmOperationOccurrenceNotification',
                    'VnfIdentifierCreationNotification',
                    'VnfIdentifierDeletionNotification']
            }
        },
        'operationTypes': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'INSTANTIATE',
                    'SCALE',
                    'SCALE_TO_LEVEL',
                    'CHANGE_FLAVOUR',
                    'TERMINATE',
                    'HEAL',
                    'OPERATE',
                    'CHANGE_EXT_CONN',
                    'MODIFY_INFO']
            }
        },
        'operationStates': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': [
                    'STARTING',
                    'PROCESSING',
                    'COMPLETED',
                    'FAILED_TEMP',
                    'FAILED',
                    'ROLLING_BACK',
                    'ROLLED_BACK']
            }
        }
    }
}

create = {
    'type': 'object',
    'properties': {
        'vnfdId': parameter_types.uuid,
        'vnfInstanceName': parameter_types.name_allow_zero_min_length,
        'vnfInstanceDescription': parameter_types.description,
        'metadata': parameter_types.keyvalue_pairs,
    },
    'required': ['vnfdId'],
    'additionalProperties': True,
}

instantiate = {
    'type': 'object',
    'properties': {
        'flavourId': {'type': 'string', 'maxLength': 255},
        'instantiationLevelId': {'type': 'string', 'maxLength': 255},
        'extVirtualLinks': _extVirtualLinkData,
        'extManagedVirtualLinks': _extManagedVirtualLinkData,
        'vimConnectionInfo': _vimConnectionInfo,
        'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['flavourId'],
    'additionalProperties': True,
}

terminate = {
    'type': 'object',
    'properties': {
        'terminationType': {'type': 'string',
                            'enum': ['FORCEFUL', 'GRACEFUL']},
        'gracefulTerminationTimeout': {'type': 'integer', 'minimum': 0},
        'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['terminationType'],
    'additionalProperties': True,
}

heal = {
    'type': 'object',
    'properties': {
        'cause': {'type': 'string', 'maxLength': 255},
        'vnfcInstanceId': {
            'type': 'array',
            "items": {
                "type": "string",
                'format': 'uuid'
            }
        }

    },
    'additionalProperties': True,
}

register_subscription = {
    'type': 'object',
    'properties': {
        'filter': _lifecycle_change_notifications_filter,
        'callbackUri': {'type': 'string', 'maxLength': 255},
        'authentication': parameter_types.keyvalue_pairs,
    },
    'required': ['callbackUri'],
    'additionalProperties': True,
}

update = {
    'type': 'object',
    'properties': {
        'vnfdId': parameter_types.uuid,
        'vnfInstanceName': parameter_types.name_allow_zero_min_length,
        'vnfInstanceDescription': parameter_types.description,
        'vnfPkgId': parameter_types.uuid,
        'metadata': parameter_types.keyvalue_pairs,
        'vimConnectionInfo': _vimConnectionInfo,
    },
    'additionalProperties': True,
}

scale = {
    'type': 'object',
    'properties': {
        'type': {'type': 'string',
                 'enum': ['SCALE_OUT', 'SCALE_IN']},
        'aspectId': {'type': 'string'},
        'numberOfSteps': {'type': 'integer'},
        'additionalParams': parameter_types.keyvalue_pairs
    },
    'required': ['type', 'aspectId'],
    'additionalProperties': True,
}

change_ext_conn = {
    'type': 'object',
    'properties': {
            'extVirtualLinks': _extVirtualLinkData,
            'vimConnectionInfo': _vimConnectionInfo,
            'additionalParams': parameter_types.keyvalue_pairs,
    },
    'required': ['extVirtualLinks'],
    'additionalProperties': True,
}

cancel = {
    'type': 'object',
    'properties': {
        'cancelMode': {'type': 'string',
                       'enum': ['FORCEFUL', 'GRACEFUL']},
    },
    'required': ['cancelMode'],
    'additionalProperties': True,
}



# Copyright (c) 2020 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tacker import context
from tacker import objects
from tacker.objects import fields
from tacker.tests.unit import base
from tacker.tests import uuidsentinel


def get_instantiated_info_dict_for_ext_links_and_flavour_id():
    return {"instantiated_vnf_info": {
        "additional_params": {'key1': 'value1'},
        "vnfc_resource_info": [{
            "vdu_id": "VDU1",
            "vnfc_cp_info": [{
                "cp_protocol_info": [{
                    "ip_over_ethernet": {
                        "mac_address": "fa:16:3e:0d:6f:71"},
                    "layer_protocol": "IP_OVER_ETHERNET"}],
                "vnf_ext_cp_id": None,
                "cpd_id": "CP1",
                "vnf_link_port_id": "a41417d8-5cba-4256-aa4a-c68b79bb2dc2",
                "id": "350ab5d7-7cf0-4d24-aa2f-ec7e51149d56"}, {
                "cp_protocol_info": [{
                    "layer_protocol": "IP_OVER_ETHERNET"}],
                "vnf_ext_cp_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe",
                "cpd_id": "CP2",
                "vnf_link_port_id": "b5f67448-87ee-4250-ad8b-514bc693f6c4",
                "id": "2cab54a3-8ac5-4338-b586-b12e4ff1a89f"}, {
                "vnf_link_port_id": "124f1923-9de1-4729-8d4e-bc0613066dc4",
                "id": "b3341e7b-20bb-4ae5-8be9-b00cf79a9395",
                "vnf_ext_cp_id": None,
                "cpd_id": "CP3"}],
            "compute_resource": {
                "vim_level_resource_type": "OS::Nova::Server",
                "resource_id": "c83ac219-376a-4016-b8b2-50c3707f71c4"},
            "id": "0688be82-05a6-4b00-b869-d4719baf4f42",
            "storage_resource_ids": [],
            "metadata": {}}],
        "ext_virtual_link_info": [{
            "resource_handle": {
                "vim_level_resource_type": None,
                "resource_id": "7ec01a11-e584-404a-88bd-39a56b63e29c"},
            "id": "net0"}, {
            "resource_handle": {
                "vim_level_resource_type": None,
                "resource_id": "dc67ee99-e963-44e2-a152-f0fb492eae76"},
            "id": "external_network",
            "ext_link_ports": [{
                "cp_instance_id": "f47a9e33-b31a-4290-828a-c7569c52bd0e",
                "resource_handle": {
                    "vim_level_resource_type": "LINKPORT",
                    "resource_id": "67f7e772-0d31-4087-bf4c-2576fadcbdb7"},
                "id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe"}]}],
        "vnf_virtual_link_resource_info": [{
            "vnf_link_ports": [{
                "cp_instance_id": "b3341e7b-20bb-4ae5-8be9-b00cf79a9395",
                "resource_handle": {
                    "vim_level_resource_type": "OS::Neutron::Port",
                    "resource_id": "87b567a0-783a-4c06-b0d8-cdafd40c51e3"},
                "id": "124f1923-9de1-4729-8d4e-bc0613066dc4"}],
            "network_resource": {
                "vim_level_resource_type": "OS::Neutron::Net",
                "resource_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfd"},
            "vnf_virtual_link_desc_id": "VL3",
            "id": "84ffa547-1164-4e60-863d-c7dac770f824"}, {
            "vnf_link_ports": [{
                "cp_instance_id": "2cab54a3-8ac5-4338-b586-b12e4ff1a89f",
                "resource_handle": {
                    "vim_level_resource_type": "OS::Neutron::Port",
                    "resource_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe"},
                "id": "b5f67448-87ee-4250-ad8b-514bc693f6c4"}],
            "network_resource": {
                "vim_level_resource_type": "OS::Neutron::Net",
                "resource_id": "dc67ee99-e963-44e2-a152-f0fb492eae76"},
            "vnf_virtual_link_desc_id": "external_network",
            "id": "1050cce6-27ca-40df-b66e-160113fc4826"}, {
            "vnf_link_ports": [{
                "cp_instance_id": "350ab5d7-7cf0-4d24-aa2f-ec7e51149d56",
                "resource_handle": {
                    "vim_level_resource_type": "OS::Neutron::Port",
                    "resource_id": "2bca5538-edd6-4590-ab1a-552ae7d0f81b"},
                "id": "a41417d8-5cba-4256-aa4a-c68b79bb2dc2"}],
            "network_resource": {
                "vim_level_resource_type": "OS::Neutron::Net",
                "resource_id": "7ec01a11-e584-404a-88bd-39a56b63e29c"},
            "vnf_virtual_link_desc_id": "net0",
            "id": "5ad98d4d-4b81-4b1b-a88b-1f64b66e151b"}],
        "ext_managed_virtual_link_info": [{
            "network_resource": {
                "vim_level_resource_type": "OS::Neutron::Net",
                "resource_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfd"},
            "id": "net1",
            "vnf_link_ports": [{
                "cp_instance_id": "b3341e7b-20bb-4ae5-8be9-b00cf79a9395",
                "resource_handle": {
                    "vim_level_resource_type": "OS::Neutron::Port",
                    "resource_id": "87b567a0-783a-4c06-b0d8-cdafd40c51e3"},
                "id": "124f1923-9de1-4729-8d4e-bc0613066dc4"}],
            "vnf_virtual_link_desc_id": "VL3"}],
        "flavour_id": "simple",
        "vnf_state": "STARTED",
        "ext_cp_info": [{
            "cp_protocol_info": [{
                "ip_over_ethernet": {
                    "mac_address": "fa:16:3e:0d:6f:71"},
                "layer_protocol": "IP_OVER_ETHERNET"}],
            "cpd_id": "CP1",
            "id": "19f0aa71-9376-43dc-8e13-5e76e4bdc8bf",
            "ext_link_port_id": None,
            "associated_vnfc_cp_id": "dc67ee99-e963-44e2-a152-f0fb492eae76"}, {
            "cp_protocol_info": [{
                "layer_protocol": "IP_OVER_ETHERNET"}],
            "cpd_id": "CP2",
            "id": "f47a9e33-b31a-4290-828a-c7569c52bd0e",
            "ext_link_port_id": None,
            "associated_vnfc_cp_id": "7ec01a11-e584-404a-88bd-39a56b63e29c"}]
    }
    }


class InstantiateVnfRequestTestCase(base.TestCase):

    def setUp(self):
        super(InstantiateVnfRequestTestCase, self).setUp()
        self.context = context.get_admin_context()

    def _create_vim_connection_info(self, **kwargs):
        vim_connection_info = objects.VimConnectionInfo(**kwargs)
        return [vim_connection_info]

    def _create_vnf_instance(self, vim_connection_info=None):
        vnf_instance = objects.VnfInstance(
            context=self.context,
            vnf_instance_name="Sample vnf instance name",
            vnf_instance_description="Sample vnf instance description",
            vnfd_id=uuidsentinel.vnfd_id,
            instantiation_state=fields.VnfInstanceState.INSTANTIATED,
            vnf_provider='Company',
            vnf_product_name='Sample VNF',
            vnf_software_version='1.0',
            vnfd_version='1.0',
            tenant_id=uuidsentinel.tenant_id)

        if vim_connection_info:
            vnf_instance.vim_connection_info = vim_connection_info

        return vnf_instance

    def test_from_vnf_instance_with_flavour(self):
        """Map flavour id"""

        inst_vnf_request = objects.InstantiateVnfRequest(flavour_id="simple")

        instantiated_vnf_info = objects.InstantiatedVnfInfo(
            flavour_id="simple")

        vnf_instance = self._create_vnf_instance()
        vnf_instance.instantiated_vnf_info = instantiated_vnf_info

        inst_vnf_request_actual = objects.InstantiateVnfRequest.\
            from_vnf_instance(vnf_instance)

        self.compare_obj(inst_vnf_request, inst_vnf_request_actual)

    def test_from_vnf_instance_with_flavour_and_instantiation_level(self):
        """Map flavour id and instantiation level"""

        inst_vnf_request = objects.InstantiateVnfRequest(flavour_id="simple",
            instantiation_level_id="instantiation_level_1")

        instantiated_vnf_info = objects.InstantiatedVnfInfo(
            flavour_id="simple",
            instantiation_level_id="instantiation_level_1")

        vnf_instance = self._create_vnf_instance()
        vnf_instance.instantiated_vnf_info = instantiated_vnf_info

        inst_vnf_request_actual = objects.InstantiateVnfRequest.\
            from_vnf_instance(vnf_instance)

        self.compare_obj(inst_vnf_request, inst_vnf_request_actual)

    def test_from_vnf_instance_with_ext_vl_and_ext_managed_vl(self):
        """Map external and internal network information.

        Map following information:
        a) ext_virtual_link_info
        b) ext_managed_virtual_link_info
        """

        ext_vl_info = [{
            "ext_cps": [{
                "cp_config": [{
                    "cp_protocol_data": [{
                        "ip_over_ethernet": {
                            "mac_address": "fa:16:3e:0d:6f:71"},
                        "layer_protocol": "IP_OVER_ETHERNET"}]}],
                "cpd_id": "CP1"}],
            "id": "net0",
            "resource_id": "7ec01a11-e584-404a-88bd-39a56b63e29c"}, {
            "ext_cps": [{
                "cp_config": [{
                    "cp_protocol_data": [{
                        "layer_protocol": "IP_OVER_ETHERNET"}],
                    "link_port_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe"}],
                "cpd_id": "CP2"}],
            "ext_link_ports": [{
                "id": "413f4e46-21cf-41b1-be0f-de8d23f76cfe",
                "resource_handle": {
                    "resource_id": "67f7e772-0d31-4087-bf4c-2576fadcbdb7",
                    "vim_level_resource_type": "LINKPORT"}}],
            "id": "external_network",
            "resource_id": "dc67ee99-e963-44e2-a152-f0fb492eae76"}]

        ext_mg_vl = {"id": "net1",
            "resource_id": "413f4e46-21cf-41b1-be0f-de8d23f76cfd",
            "vnf_virtual_link_desc_id": "VL3"}

        vim_connection_info = [{
            "id": "6b0ff598-60d6-49b4-a907-a1111de52d92",
            "vim_id": uuidsentinel.vim_id,
            "vim_type": "ETSINFV.OPENSTACK_KEYSTONE.v_2",
            "access_info": {}}]

        vnf_info_data = {'ext_managed_virtual_links': [ext_mg_vl],
            'ext_virtual_links': ext_vl_info,
            'vim_connection_info': vim_connection_info,
            'flavour_id': 'simple',
            'additional_params': {'key1': 'value1'}}

        inst_vnf_request = objects.InstantiateVnfRequest.obj_from_primitive(
            vnf_info_data, self.context)

        response = get_instantiated_info_dict_for_ext_links_and_flavour_id()

        instantiated_vnf_info_dict = response.get('instantiated_vnf_info')

        instantiated_vnf_info = objects.InstantiatedVnfInfo.obj_from_primitive(
            instantiated_vnf_info_dict, self.context)

        vnf_instance = self._create_vnf_instance(
            self._create_vim_connection_info(**vim_connection_info[0]))
        vnf_instance.instantiated_vnf_info = instantiated_vnf_info

        inst_vnf_req_actual = objects.InstantiateVnfRequest.from_vnf_instance(
            vnf_instance)

        self.assertEqual(inst_vnf_request.flavour_id,
            inst_vnf_req_actual.flavour_id)

        self.assertEqual(inst_vnf_request.instantiation_level_id,
            inst_vnf_req_actual.instantiation_level_id)

        self.assertEqual(inst_vnf_request.additional_params,
            inst_vnf_req_actual.additional_params)

        for expected, actual in zip(inst_vnf_request.ext_managed_virtual_links,
                inst_vnf_req_actual.ext_managed_virtual_links):
            self.compare_obj(expected, actual)

        for expected, actual in zip(inst_vnf_request.ext_virtual_links,
                inst_vnf_req_actual.ext_virtual_links):
            self.assertEqual(expected.id, actual.id)
            self.assertEqual(expected.resource_id, actual.resource_id)

        for expected, actual in zip(inst_vnf_request.vim_connection_info,
                inst_vnf_req_actual.vim_connection_info):
            self.assertEqual(expected.id, actual.id)
            self.assertEqual(expected.vim_id, actual.vim_id)
            self.assertEqual(expected.vim_type, actual.vim_type)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import routes
import webob

import oslo_i18n as i18n
from oslo_log import log as logging

from tacker.common import exceptions as common_ex
from tacker import wsgi

from tacker.sol_refactored.api import api_version
from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import exceptions as sol_ex


LOG = logging.getLogger(__name__)


class SolResponse(object):

    # SOL013 4.2.3 Response header field
    allowed_headers = ['version', 'location', 'content_type',
            'www_authenticate', 'accept_ranges', 'content_range',
            'retry_after', 'link']

    def __init__(self, status, body, **kwargs):
        self.status = status
        self.body = body
        self.headers = {}
        for hdr in self.allowed_headers:
            if hdr in kwargs:
                self.headers[hdr] = kwargs[hdr]
        self.headers.setdefault('version', api_version.CURRENT_VERSION)
        self.headers.setdefault('accept-ranges', 'none')

    def serialize(self, request, content_type):
        self.headers.setdefault('content_type', content_type)
        content_type = self.headers['content_type']
        if self.body is None:
            body = None
        elif content_type == 'text/plain':
            body = self.body
        elif content_type == 'application/zip':
            body = self.body
        else:  # 'application/json'
            serializer = wsgi.JSONDictSerializer()
            body = serializer.serialize(self.body)
            if len(body) > config.CONF.v2_vnfm.max_content_length:
                raise sol_ex.ResponseTooBig(
                    size=config.CONF.v2_vnfm.max_content_length)
        response = webob.Response(body=body)
        response.status_int = self.status
        for hdr, val in self.headers.items():
            response.headers[hdr.replace('_', '-')] = val
        return response


class SolErrorResponse(SolResponse):

    def __init__(self, ex, req):
        user_locale = req.best_match_language()
        problem_details = {}
        if isinstance(ex, sol_ex.SolException):
            problem_details = ex.make_problem_details()
            # translate detail
            detail = i18n.translate(problem_details['detail'], user_locale)
            problem_details['detail'] = detail
        elif isinstance(ex, (common_ex.TackerException,
                             webob.exc.HTTPException)):
            LOG.warning("legacy Exception used. Use SolException instead.")
            problem_details['status'] = ex.code
            problem_details['detail'] = str(ex)
        else:
            # program bug. it occurs only under development.
            LOG.exception("Unknown error")
            problem_details['status'] = 500
            problem_details['detail'] = str(ex)

        super(SolErrorResponse, self).__init__(problem_details['status'],
                                               problem_details)


class SolResource(wsgi.Application):

    def __init__(self, controller, policy_name=None):
        self.controller = controller
        self.policy_name = policy_name
        self.deserializer = wsgi.RequestDeserializer()

    @webob.dec.wsgify(RequestClass=wsgi.Request)
    def __call__(self, request):
        LOG.info("%(method)s %(url)s", {"method": request.method,
                                        "url": request.url})
        try:
            action, args, accept = self.deserializer.deserialize(request)
            self.check_api_version(request)
            self.check_policy(request, action)
            result = self.dispatch(request, action, args)
            response = result.serialize(request, accept)
        except Exception as ex:
            result = SolErrorResponse(ex, request)
            try:
                response = result.serialize(request,
                                            'application/problem+json')
            except Exception:
                LOG.exception("Unknown error")
                return webob.exc.HTTPBadRequest(explanation="Unknown error")

        LOG.info("%(url)s returned with HTTP %(status)d",
                 {"url": request.url, "status": response.status_int})

        return response

    def check_api_version(self, request):
        # check and set api_version
        ver = request.headers.get("Version")
        if ver is None:
            LOG.info("Version missing")
            raise sol_ex.APIVersionMissing()
        request.context.api_version = api_version.APIVersion(ver)

    def check_policy(self, request, action):
        if self.policy_name is None:
            return
        if action == 'reject':
            return
        request.context.can(self.policy_name.format(action))

    def dispatch(self, request, action, action_args):
        controller_method = getattr(self.controller, action)
        return controller_method(request=request, **action_args)


class SolAPIRouter(wsgi.Router):

    controller = None
    route_list = {}

    def __init__(self):
        super(SolAPIRouter, self).__init__(routes.Mapper())

    def _setup_routes(self, mapper):
        for path, methods in self.route_list:
            self._setup_route(mapper, path, methods)

    def _setup_route(self, mapper, path, methods):

        for method, action in methods.items():
            mapper.connect(path,
                           controller=self.controller,
                           action=action,
                           conditions={'method': [method]})

        all_methods = ['HEAD', 'GET', 'POST', 'PUT', 'PATCH', 'DELETE']
        missing_methods = [m for m in all_methods if m not in methods]
        if missing_methods:
            mapper.connect(path,
                           controller=self.controller,
                           action='reject',
                           conditions={'method': missing_methods})


class SolAPIController(object):

    def reject(self, request, **kwargs):
        raise sol_ex.MethodNotAllowed(method=request.method)



# Copyright (c) 2012 OpenStack Foundation
# Copyright (C) 2022 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import os
from oslo_log import log as logging
import unittest

from tacker.agent.linux import utils
import tacker.privileged.linux_cmd
from tacker.tests import base


LOG = logging.getLogger(__name__)

# Use env 'PWD' to check if tests are run on zuul because we cannot run the
# tests require root privileges. Skip them on zuul, but still run on
# localhost to test privsep features.
_PWD = os.environ['PWD']
_PWD_ZUUL = "/home/zuul/src/opendev.org/openstack/tacker"


class PrivsepTest(base.BaseTestCase):
    """Simple unit test to test the basic privsep mechanism

    Essentially hello-world. Just run a command as root and check that
    it actually *did* run as root.
    """

    def setUp(self):
        super(PrivsepTest, self).setUp()

    @unittest.skipIf(_PWD == _PWD_ZUUL, "Failed on zuul for root priv")
    def test_privsep_ls(self):
        """Run ls with root privilege

        This ls command is expected to be run on `/`.
        """

        ls = tacker.privileged.linux_cmd.ls()
        # The result is a series of dirs on '/' and separated with '\n' like
        # as 'bin\nboot\ndev\netc\n...'.
        res = ls[0].split('\n')

        # 'boot' dir must be under '/'.
        self.assertIn('boot', res)

    @unittest.skipIf(_PWD == _PWD_ZUUL, "Failed on zuul for root priv")
    def test_privsep_pwd(self):
        """Run pwd with root privilege

        This ls command is expected to be run on `/`.
        """
        res = tacker.privileged.linux_cmd.pwd()[0]
        self.assertEqual('/\n', res)

    @unittest.skipIf(_PWD == _PWD_ZUUL, "Failed on zuul for root priv")
    def test_rootwrap(self):
        """Confirm a command can be run with tacker-rootwrap

        pwd is used as a harmless command in this test and defined in
        '/etc/tacker/rootwrap.d/tacker.filters' as a CommandFilter.
        """

        root_helper = ["sudo", "tacker-rootwrap",
                       "/etc/tacker/rootwrap.conf"]
        cmd = "pwd"

        actual = utils.execute(root_helper + [cmd])
        expected = utils.execute([cmd])

        self.assertEqual(expected, actual)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker import objects
from tacker.objects import base
from tacker.objects import fields


@base.TackerObjectRegistry.register
class Grant(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnf_instance_id': fields.StringField(nullable=False),
        'vnf_lcm_op_occ_id': fields.StringField(nullable=False),
        'vim_connections': fields.ListOfObjectsField(
            'VimConnectionInfo', nullable=True, default=[]),
        'zones': fields.ListOfObjectsField(
            'ZoneInfo', nullable=True, default=[]),
        'add_resources': fields.ListOfObjectsField(
            'GrantInfo', nullable=True, default=[]),
        'remove_resources': fields.ListOfObjectsField(
            'GrantInfo', nullable=True, default=[]),
        'update_resources': fields.ListOfObjectsField(
            'GrantInfo', nullable=True, default=[]),
        'vim_assets': fields.ObjectField(
            'VimAssets', nullable=True),
        'ext_virtual_links': fields.ListOfObjectsField(
            'ExtVirtualLinkData', nullable=True, default=[]),
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_grant = super(
                Grant, cls).obj_from_primitive(primitive, context)
        else:
            if 'vim_connections' in primitive.keys():
                obj_data = [objects.VimConnectionInfo._from_dict(
                    vim_conn) for vim_conn in primitive.get(
                    'vim_connections', [])]
                primitive.update({'vim_connections': obj_data})

            if 'zones' in primitive.keys():
                obj_data = [ZoneInfo._from_dict(
                    zone) for zone in primitive.get(
                    'zones', [])]
                primitive.update({'zones': obj_data})

            if 'add_resources' in primitive.keys():
                obj_data = [GrantInfo._from_dict(
                    add_rsc) for add_rsc in primitive.get(
                    'add_resources', [])]
                primitive.update({'add_resources': obj_data})
            if 'remove_resources' in primitive.keys():
                obj_data = [GrantInfo._from_dict(
                    remove_rsc) for remove_rsc in primitive.get(
                    'remove_resources', [])]
                primitive.update({'remove_resources': obj_data})
            if 'update_resources' in primitive.keys():
                obj_data = [GrantInfo._from_dict(
                    update_rsc) for update_rsc in primitive.get(
                    'update_resources', [])]
                primitive.update({'update_resources': obj_data})
            if 'vim_assets' in primitive.keys():
                obj_data = VimAssets.obj_from_primitive(
                    primitive.get('vim_assets'), context)
                primitive.update({'vim_assets': obj_data})
            if 'ext_virtual_links' in primitive.keys():
                obj_data = [objects.ExtVirtualLinkData.obj_from_primitive(
                    ext_vir_link, context) for ext_vir_link in primitive.get(
                    'ext_virtual_links', [])]
                primitive.update({'ext_virtual_links': obj_data})
            obj_grant = Grant._from_dict(primitive)

        return obj_grant

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vnf_instance_id = data_dict.get('vnf_instance_id')
        vnf_lcm_op_occ_id = data_dict.get('vnf_lcm_op_occ_id')
        vim_connections = data_dict.get('vim_connections', [])
        zones = data_dict.get('zones', [])
        add_resources = data_dict.get('add_resources', [])
        remove_resources = data_dict.get('remove_resources', [])
        update_resources = data_dict.get('update_resources', [])
        vim_assets = data_dict.get('vim_assets')
        ext_virtual_links = data_dict.get('ext_virtual_links', [])

        obj = cls(
            id=id,
            vnf_instance_id=vnf_instance_id,
            vnf_lcm_op_occ_id=vnf_lcm_op_occ_id,
            vim_connections=vim_connections,
            zones=zones,
            add_resources=add_resources,
            remove_resources=remove_resources,
            update_resources=update_resources,
            vim_assets=vim_assets,
            ext_virtual_links=ext_virtual_links)
        return obj


@base.TackerObjectRegistry.register
class ZoneInfo(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'zone_id': fields.StringField(nullable=False),
        'vim_connection_id': fields.StringField(nullable=True)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_zone_info = super(
                ZoneInfo, cls).obj_from_primitive(primitive, context)
        else:
            obj_zone_info = ZoneInfo._from_dict(primitive)

        return obj_zone_info

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        zone_id = data_dict.get('zone_id')
        vim_connection_id = data_dict.get('vim_connection_id')

        obj = cls(
            id=id,
            zone_id=zone_id,
            vim_connection_id=vim_connection_id)
        return obj


@base.TackerObjectRegistry.register
class GrantInfo(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'resource_definition_id': fields.StringField(nullable=False),
        'vim_connection_id': fields.StringField(nullable=True),
        'zone_id': fields.StringField(nullable=True)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_grant_info = super(
                GrantInfo, cls).obj_from_primitive(primitive, context)
        else:
            obj_grant_info = GrantInfo._from_dict(primitive)

        return obj_grant_info

    @classmethod
    def _from_dict(cls, data_dict):
        resource_definition_id = data_dict.get('resource_definition_id')
        vim_connection_id = data_dict.get('vim_connection_id')
        zone_id = data_dict.get('zone_id')

        obj = cls(
            resource_definition_id=resource_definition_id,
            vim_connection_id=vim_connection_id,
            zone_id=zone_id)
        return obj


@base.TackerObjectRegistry.register
class VimAssets(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'compute_resource_flavours': fields.ListOfObjectsField(
            'VimComputeResourceFlavour', nullable=True, default=[]),
        'software_images': fields.ListOfObjectsField(
            'VimSoftwareImage', nullable=True, default=[])
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_vim_assets = super(
                VimAssets, cls).obj_from_primitive(primitive, context)
        else:
            if 'compute_resource_flavours' in primitive.keys():
                obj_data = [VimComputeResourceFlavour._from_dict(
                    flavour) for flavour in primitive.get(
                    'compute_resource_flavours', [])]
                primitive.update({'compute_resource_flavours': obj_data})

            if 'software_images' in primitive.keys():
                obj_data = [VimSoftwareImage._from_dict(
                    img) for img in primitive.get(
                    'software_images', [])]
                primitive.update({'software_images': obj_data})
            obj_vim_assets = VimAssets._from_dict(primitive)

        return obj_vim_assets

    @classmethod
    def _from_dict(cls, data_dict):
        compute_resource_flavours = data_dict.get(
            'compute_resource_flavours', [])
        software_images = data_dict.get('software_images', [])

        obj = cls(
            compute_resource_flavours=compute_resource_flavours,
            software_images=software_images)
        return obj


@base.TackerObjectRegistry.register
class VimComputeResourceFlavour(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_id': fields.StringField(nullable=True),
        'vnfd_virtual_compute_desc_id': fields.StringField(nullable=False),
        'vim_flavour_id': fields.StringField(nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_flavour = super(
                VimComputeResourceFlavour,
                cls).obj_from_primitive(
                primitive,
                context)
        else:
            obj_flavour = VimComputeResourceFlavour._from_dict(primitive)

        return obj_flavour

    @classmethod
    def _from_dict(cls, data_dict):
        vim_connection_id = data_dict.get('vim_connection_id')
        vnfd_virtual_compute_desc_id = data_dict.get(
            'vnfd_virtual_compute_desc_id')
        vim_flavour_id = data_dict.get('vim_flavour_id')

        obj = cls(
            vim_connection_id=vim_connection_id,
            vnfd_virtual_compute_desc_id=vnfd_virtual_compute_desc_id,
            vim_flavour_id=vim_flavour_id)
        return obj


@base.TackerObjectRegistry.register
class VimSoftwareImage(base.TackerObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_id': fields.StringField(nullable=True),
        'vnfd_software_image_id': fields.StringField(nullable=False),
        'vim_software_image_id': fields.StringField(nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            obj_img = super(
                VimSoftwareImage, cls).obj_from_primitive(primitive, context)
        else:
            obj_img = VimSoftwareImage._from_dict(primitive)

        return obj_img

    @classmethod
    def _from_dict(cls, data_dict):
        vim_connection_id = data_dict.get('vim_connection_id')
        vnfd_software_image_id = data_dict.get('vnfd_software_image_id')
        vim_software_image_id = data_dict.get('vim_software_image_id')

        obj = cls(
            vim_connection_id=vim_connection_id,
            vnfd_software_image_id=vnfd_software_image_id,
            vim_software_image_id=vim_software_image_id)
        return obj
